@@ -4,7 +4,7 @@
   iconv.c -
 
   $Author: shyouhei $
-  $Date: 2008-06-15 19:20:39 +0900 (Sun, 15 Jun 2008) $
+  $Date: 2009-11-24 16:18:40 +0900 (Tue, 24 Nov 2009) $
   created at: Wed Dec  1 20:28:09 JST 1999
 
   All the files in this distribution are covered under the Ruby's
@@ -43,7 +43,11 @@
  * 
  * == Examples
  *
- * 1. Instantiate a new Iconv and use method Iconv#iconv.
+ * 1. Simple conversion between two charsets.
+ *
+ *      converted_text = Iconv.conv('iso-8859-15', 'utf-8', text)
+ *
+ * 2. Instantiate a new Iconv and use method Iconv#iconv.
  *
  *      cd = Iconv.new(to, from)
  *      begin
@@ -53,20 +57,16 @@
  *        cd.close
  *      end
  *
- * 2. Invoke Iconv.open with a block.
+ * 3. Invoke Iconv.open with a block.
  *
  *      Iconv.open(to, from) do |cd|
  *        input.each { |s| output << cd.iconv(s) }
  *        output << cd.iconv(nil)
  *      end
  *
- * 3. Shorthand for (2).
+ * 4. Shorthand for (3).
  *
  *      Iconv.iconv(to, from, *input.to_a)
- *
- * 4. Simple conversion between two charsets.
- *
- *      converted_text = Iconv.new('iso-8859-15', 'utf-8').iconv(text)
  */
 
 /* Invalid value for iconv_t is -1 but 0 for VALUE, I hope VALUE is
@@ -92,6 +92,7 @@ static VALUE rb_eIconvBrokenLibrary;
 
 static ID rb_success, rb_failed;
 static VALUE iconv_fail _((VALUE error, VALUE success, VALUE failed, struct iconv_env_t* env, const char *mesg));
+static VALUE iconv_fail_retry _((VALUE error, VALUE success, VALUE failed, struct iconv_env_t* env, const char *mesg));
 static VALUE iconv_failure_initialize _((VALUE error, VALUE mesg, VALUE success, VALUE failed));
 static VALUE iconv_failure_success _((VALUE self));
 static VALUE iconv_failure_failed _((VALUE self));
@@ -101,7 +102,7 @@ static void iconv_dfree _((void *cd));
 static VALUE iconv_free _((VALUE cd));
 static VALUE iconv_try _((iconv_t cd, const char **inptr, size_t *inlen, char **outptr, size_t *outlen));
 static VALUE rb_str_derive _((VALUE str, const char* ptr, int len));
-static VALUE iconv_convert _((iconv_t cd, VALUE str, int start, int length, struct iconv_env_t* env));
+static VALUE iconv_convert _((iconv_t cd, VALUE str, long start, long length, struct iconv_env_t* env));
 static VALUE iconv_s_allocate _((VALUE klass));
 static VALUE iconv_initialize _((VALUE self, VALUE to, VALUE from));
 static VALUE iconv_s_open _((VALUE self, VALUE to, VALUE from));
@@ -170,7 +171,7 @@ iconv_create
 	}
 	if (cd == (iconv_t)-1) {
 	    int inval = errno == EINVAL;
-	    char *s = inval ? "invalid encoding " : "iconv";
+	    const char *s = inval ? "invalid encoding " : "iconv";
 	    volatile VALUE msg = rb_str_new(0, strlen(s) + RSTRING(to)->len +
 					    RSTRING(from)->len + 8);
 
@@ -179,8 +180,8 @@ iconv_create
 	    s = RSTRING(msg)->ptr;
 	    RSTRING(msg)->len = strlen(s);
 	    if (!inval) rb_sys_fail(s);
-	    iconv_fail(rb_eIconvInvalidEncoding,
-		       Qnil, rb_ary_new3(2, to, from), NULL, s);
+	    rb_exc_raise(iconv_fail(rb_eIconvInvalidEncoding, Qnil,
+				    rb_ary_new3(2, to, from), NULL, s));
 	}
     }
 
@@ -328,7 +329,13 @@ iconv_fail
 	    args[2] = rb_ary_new4(env->argc, env->argv);
 	}
     }
-    error = rb_class_new_instance(3, args, error);
+    return rb_class_new_instance(3, args, error);
+}
+
+static VALUE
+iconv_fail_retry(VALUE error, VALUE success, VALUE failed, struct iconv_env_t* env, const char *mesg)
+{
+    error = iconv_fail(error, success, failed, env, mesg);
     if (!rb_block_given_p()) rb_exc_raise(error);
     ruby_errinfo = error;
     return rb_yield(failed);
@@ -362,13 +369,13 @@ rb_str_derive
 static VALUE
 iconv_convert
 #ifdef HAVE_PROTOTYPES
-    (iconv_t cd, VALUE str, int start, int length, struct iconv_env_t* env)
+    (iconv_t cd, VALUE str, long start, long length, struct iconv_env_t* env)
 #else /* HAVE_PROTOTYPES */
     (cd, str, start, length, env)
     iconv_t cd;
     VALUE str;
-    int start;
-    int length;
+    long start;
+    long length;
     struct iconv_env_t *env;
 #endif /* HAVE_PROTOTYPES */
 {
@@ -394,7 +401,7 @@ iconv_convert
 	error = iconv_try(cd, &inptr, &inlen, &outptr, &outlen);
 	if (RTEST(error)) {
 	    unsigned int i;
-	    rescue = iconv_fail(error, Qnil, Qnil, env, 0);
+	    rescue = iconv_fail_retry(error, Qnil, Qnil, env, 0);
 	    if (TYPE(rescue) == T_ARRAY) {
 		str = RARRAY(rescue)->len > 0 ? RARRAY(rescue)->ptr[0] : Qnil;
 	    }
@@ -417,17 +424,9 @@ iconv_convert
 	slen = RSTRING(str)->len;
 	inptr = RSTRING(str)->ptr;
 
-	if (start < 0 ? (start += slen) < 0 : start >= slen)
-	    length = 0;
-	else if (length < 0 && (length += slen + 1) < 0)
-	    length = 0;
-	else if ((length -= start) < 0)
-	    length = 0;
-	else {
-	    inptr += start;
-	    if (length > slen)
-		length = slen;
-	}
+	inptr += start;
+	if (length < 0 || length > start + slen)
+	    length = slen - start;
     }
     instart = inptr;
     inlen = length;
@@ -481,7 +480,7 @@ iconv_convert
 	    else if (inptr > instart)
 		rb_str_cat(ret, instart, inptr - instart);
 	    str = rb_str_derive(str, inptr, inlen);
-	    rescue = iconv_fail(error, ret, str, env, errmsg);
+	    rescue = iconv_fail_retry(error, ret, str, env, errmsg);
 	    if (TYPE(rescue) == T_ARRAY) {
 		if ((len = RARRAY(rescue)->len) > 0)
 		    rb_str_concat(ret, RARRAY(rescue)->ptr[0]);
@@ -609,7 +608,7 @@ iconv_s_convert
 }
 
 /*
- * Document-method: iconv
+ * Document-method: Iconv::iconv
  * call-seq: Iconv.iconv(to, from, *strs)
  *
  * Shorthand for
@@ -718,7 +717,7 @@ iconv_finish
 }
 
 /*
- * Document-method: iconv
+ * Document-method: Iconv#iconv
  * call-seq: iconv(str, start=0, length=-1)
  *
  * Converts string and returns the result.
@@ -757,14 +756,22 @@ iconv_iconv
 {
     VALUE str, n1, n2;
     VALUE cd = check_iconv(self);
+    long start = 0, length = 0, slen = 0;
 
-    n1 = n2 = Qnil;
     rb_scan_args(argc, argv, "12", &str, &n1, &n2);
+    if (!NIL_P(str)) slen = RSTRING_LEN(StringValue(str));
+    if (argc != 2 || !RTEST(rb_range_beg_len(n1, &start, &length, slen, 0))) {
+	if (NIL_P(n1) || ((start = NUM2LONG(n1)) < 0 ? (start += slen) >= 0 : start < slen)) {
+	    if (NIL_P(n2)) {
+		length = -1;
+	    }
+	    else if ((length = NUM2LONG(n2)) >= slen - start) {
+		length = slen - start;
+	    }
+	}
+    }
 
-    return iconv_convert(VALUE2ICONV(cd), str,
-			 NIL_P(n1) ? 0 : NUM2INT(n1),
-			 NIL_P(n2) ? -1 : NUM2INT(n2),
-			 NULL);
+    return iconv_convert(VALUE2ICONV(cd), str, start, length, NULL);
 }
 
 /*
@@ -828,7 +835,7 @@ iconv_failure_inspect
     VALUE self;
 #endif /* HAVE_PROTOTYPES */
 {
-    char *cname = rb_class2name(CLASS_OF(self));
+    const char *cname = rb_class2name(CLASS_OF(self));
     VALUE success = rb_attr_get(self, rb_success);
     VALUE failed = rb_attr_get(self, rb_failed);
     VALUE str = rb_str_buf_cat2(rb_str_new2("#<"), cname);
