@@ -3,7 +3,7 @@
   stringio.c -
 
   $Author: shyouhei $
-  $Date: 2009-02-08 21:39:52 -0700 (Sun, 08 Feb 2009) $
+  $Date: 2010-11-24 16:38:34 +0900 (Wed, 24 Nov 2010) $
   $RoughId: stringio.c,v 1.13 2002/03/14 03:24:18 nobu Exp $
   created at: Tue Feb 19 04:10:38 JST 2002
 
@@ -14,6 +14,7 @@
 
 #include "ruby.h"
 #include "rubyio.h"
+#include "re.h"
 #if defined(HAVE_FCNTL_H) || defined(_WIN32)
 #include <fcntl.h>
 #elif defined(HAVE_SYS_FCNTL_H)
@@ -162,6 +163,7 @@ static VALUE strio_rewind _((VALUE));
 static VALUE strio_seek _((int, VALUE *, VALUE));
 static VALUE strio_get_sync _((VALUE));
 static VALUE strio_each_byte _((VALUE));
+static VALUE strio_each_char _((VALUE));
 static VALUE strio_getc _((VALUE));
 static VALUE strio_ungetc _((VALUE, VALUE));
 static VALUE strio_readchar _((VALUE));
@@ -267,6 +269,8 @@ strio_init(argc, argv, ptr)
 	break;
     }
     ptr->string = string;
+    ptr->pos = 0;
+    ptr->lineno = 0;
 }
 
 static VALUE
@@ -657,6 +661,9 @@ strio_seek(argc, argv, self)
 
     rb_scan_args(argc, argv, "11", NULL, &whence);
     offset = NUM2LONG(argv[0]);
+    if (CLOSED(ptr)) {
+	rb_raise(rb_eIOError, "closed stream");
+    }
     switch (NIL_P(whence) ? 0 : NUM2LONG(whence)) {
       case 0:
 	break;
@@ -667,7 +674,7 @@ strio_seek(argc, argv, self)
 	offset += RSTRING(ptr->string)->len;
 	break;
       default:
-	rb_raise(rb_eArgError, "invalid whence %ld", NUM2LONG(whence));
+	error_inval("invalid whence");
     }
     if (offset < 0) {
 	error_inval(0);
@@ -707,11 +714,14 @@ strio_each_byte(self)
     VALUE self;
 {
     struct StringIO *ptr = readable(StringIO(self));
-    while (ptr->pos < RSTRING(ptr->string)->len) {
-	char c = RSTRING(ptr->string)->ptr[ptr->pos++];
+
+    RETURN_ENUMERATOR(self, 0, 0);
+
+    while (ptr->pos < RSTRING_LEN(ptr->string)) {
+	char c = RSTRING_PTR(ptr->string)[ptr->pos++];
 	rb_yield(CHR2FIX(c));
     }
-    return Qnil;
+    return self;
 }
 
 /*
@@ -809,6 +819,41 @@ strio_readchar(self)
     return c;
 }
 
+/*
+ * call-seq:
+ *   strio.each_char {|char| block }  -> strio
+ *
+ * See IO#each_char.
+ */
+static VALUE
+strio_each_char(self)
+    VALUE self;
+{
+    struct StringIO *sio;
+    VALUE str;
+    const char *ptr;
+    size_t len;
+
+    RETURN_ENUMERATOR(self, 0, 0);
+
+    sio = readable(StringIO(self));
+    str = sio->string;
+    ptr = RSTRING_PTR(str);
+    len = RSTRING_LEN(str);
+
+    while (sio->pos < len) {
+	int pos = sio->pos;
+	char c = ptr[pos];
+        int n = mbclen(c);
+
+	if (len < pos + n) n = len - pos;
+
+	sio->pos += n;
+        rb_yield(rb_str_substr(str, pos, n));
+    }
+    return self;
+}
+
 static void
 bm_init_skip(skip, pat, m)
      long *skip;
@@ -888,7 +933,7 @@ strio_getline(argc, argv, ptr)
 	s = p;
 	while ((p = memchr(p, '\n', e - p)) && (p != e)) {
 	    if (*++p == '\n') {
-		e = p;
+		e = p + 1;
 		break;
 	    }
 	}
@@ -977,6 +1022,8 @@ strio_each(argc, argv, self)
     struct StringIO *ptr = StringIO(self);
     VALUE line;
 
+    RETURN_ENUMERATOR(self, argc, argv);
+
     while (!NIL_P(line = strio_getline(argc, argv, readable(ptr)))) {
 	rb_yield(line);
     }
@@ -1300,11 +1347,17 @@ Init_stringio()
     rb_define_method(StringIO, "path", strio_path, 0);
 
     rb_define_method(StringIO, "each", strio_each, -1);
-    rb_define_method(StringIO, "each_byte", strio_each_byte, 0);
     rb_define_method(StringIO, "each_line", strio_each, -1);
+    rb_define_method(StringIO, "lines", strio_each, -1);
+    rb_define_method(StringIO, "each_byte", strio_each_byte, 0);
+    rb_define_method(StringIO, "bytes", strio_each_byte, 0);
+    rb_define_method(StringIO, "each_char", strio_each_char, 0);
+    rb_define_method(StringIO, "chars", strio_each_char, 0);
     rb_define_method(StringIO, "getc", strio_getc, 0);
+    rb_define_method(StringIO, "getbyte", strio_getc, 0);
     rb_define_method(StringIO, "ungetc", strio_ungetc, 1);
     rb_define_method(StringIO, "readchar", strio_readchar, 0);
+    rb_define_method(StringIO, "readbyte", strio_readchar, 0);
     rb_define_method(StringIO, "gets", strio_gets, -1);
     rb_define_method(StringIO, "readline", strio_readline, -1);
     rb_define_method(StringIO, "readlines", strio_readlines, -1);
