@@ -12,9 +12,6 @@
 #include <ruby.h>
 #include <intern.h>
 #include <rubysig.h>
-#include <node.h>
-
-enum rb_thread_status rb_thread_status _((VALUE));
 
 static VALUE rb_cMutex;
 static VALUE rb_cConditionVariable;
@@ -230,9 +227,7 @@ wake_one(List *list)
 
     waking = Qnil;
     while (list->entries && !RTEST(waking)) {
-	waking = shift_list(list);
-	if (waking == Qundef) break;
-	waking = wake_thread(waking);
+	waking = wake_thread(shift_list(list));
     }
 
     return waking;
@@ -247,29 +242,26 @@ wake_all(List *list)
     return Qnil;
 }
 
-extern int rb_thread_join _((VALUE thread, double limit));
-#define DELAY_INFTY 1E30
-
 static VALUE
-wait_list_inner(VALUE arg)
+wait_list_inner(List *list)
 {
-    push_list((List *)arg, rb_thread_current());
+    push_list(list, rb_thread_current());
     rb_thread_stop();
     return Qnil;
 }
 
 static VALUE
-wait_list_cleanup(VALUE arg)
+wait_list_cleanup(List *list)
 {
     /* cleanup in case of spurious wakeups */
-    remove_one((List *)arg, rb_thread_current());
+    remove_one(list, rb_thread_current());
     return Qnil;
 }
 
-static void
+static VALUE
 wait_list(List *list)
 {
-    rb_ensure(wait_list_inner, (VALUE)list, wait_list_cleanup, (VALUE)list);
+    return rb_ensure(wait_list_inner, (VALUE)list, wait_list_cleanup, (VALUE)list);
 }
 
 static void
@@ -398,25 +390,6 @@ rb_mutex_try_lock(VALUE self)
     return Qtrue;
 }
 
-static VALUE
-wait_mutex(VALUE arg)
-{
-    Mutex *mutex = (Mutex *)arg;
-    VALUE current = rb_thread_current();
-
-    push_list(&mutex->waiting, current);
-    do {
-	rb_thread_critical = 0;
-	rb_thread_join(mutex->owner, DELAY_INFTY);
-	rb_thread_critical = 1;
-	if (!MUTEX_LOCKED_P(mutex)) {
-	    mutex->owner = current;
-	    break;
-	}
-    } while (mutex->owner != current);
-    return Qnil;
-}
-
 /*
  * Document-method: lock
  * call-seq: lock
@@ -437,19 +410,20 @@ lock_mutex(Mutex *mutex)
 	mutex->owner = current;
     }
     else {
-	rb_ensure(wait_mutex, (VALUE)mutex, wait_list_cleanup, (VALUE)&mutex->waiting);
+	do {
+	    wait_list(&mutex->waiting);
+	    rb_thread_critical = 1;
+	    if (!MUTEX_LOCKED_P(mutex)) {
+		mutex->owner = current;
+		break;
+	    }
+	} while (mutex->owner != current);
     }
 
     rb_thread_critical = 0;
     return Qnil;
 }
 
-static VALUE
-lock_mutex_call(VALUE mutex)
-{
-    return lock_mutex((Mutex *)mutex);
-}
-
 static VALUE
 rb_mutex_lock(VALUE self)
 {
@@ -459,22 +433,6 @@ rb_mutex_lock(VALUE self)
     return self;
 }
 
-static VALUE
-relock_mutex(Mutex *mutex)
-{
-    VALUE current = rb_thread_current();
-
-    switch (rb_thread_status(current)) {
-      case THREAD_RUNNABLE:
-      case THREAD_STOPPED:
-	lock_mutex(mutex);
-	break;
-      default:
-	break;
-    }
-    return Qundef;
-}
-
 /*
  * Document-method: unlock
  *
@@ -518,12 +476,6 @@ unlock_mutex(Mutex *mutex)
     return Qtrue;
 }
 
-static VALUE
-unlock_mutex_call(VALUE mutex)
-{
-    return unlock_mutex((Mutex *)mutex);
-}
-
 static VALUE
 rb_mutex_unlock(VALUE self)
 {
@@ -677,17 +629,8 @@ rb_condvar_alloc(VALUE klass)
  *
  */
 
-static void condvar_wakeup(Mutex *mutex);
-
 static void
 wait_condvar(ConditionVariable *condvar, Mutex *mutex)
-{
-    condvar_wakeup(mutex);
-    rb_ensure(wait_list, (VALUE)&condvar->waiting, lock_mutex_call, (VALUE)mutex);
-}
-
-static void
-condvar_wakeup(Mutex *mutex)
 {
     VALUE waking;
 
@@ -700,6 +643,7 @@ condvar_wakeup(Mutex *mutex)
     if (RTEST(waking)) {
 	wake_thread(waking);
     }
+    rb_ensure(wait_list, (VALUE)&condvar->waiting, lock_mutex, (VALUE)mutex);
 }
 
 static VALUE
@@ -781,13 +725,6 @@ signal_condvar(ConditionVariable *condvar)
     }
 }
 
-static VALUE
-signal_condvar_call(VALUE condvar)
-{
-    signal_condvar((ConditionVariable *)condvar);
-    return Qundef;
-}
-
 static VALUE
 rb_condvar_signal(VALUE self)
 {
@@ -1012,16 +949,6 @@ rb_queue_num_waiting(VALUE self)
     return result;
 }
 
-static void
-wait_queue(ConditionVariable *condvar, Mutex *mutex)
-{
-    condvar_wakeup(mutex);
-    wait_list(&condvar->waiting);
-    lock_mutex(mutex);
-}
-
-static VALUE queue_pop_inner(VALUE arg);
-
 /*
  * Document-method: pop
  * call_seq: pop(non_block=false)
@@ -1037,6 +964,7 @@ rb_queue_pop(int argc, VALUE *argv, VALUE self)
 {
     Queue *queue;
     int should_block;
+    VALUE result;
     Data_Get_Struct(self, Queue, queue);
 
     if (argc == 0) {
@@ -1054,21 +982,15 @@ rb_queue_pop(int argc, VALUE *argv, VALUE self)
     }
 
     while (!queue->values.entries) {
-        wait_queue(&queue->value_available, &queue->mutex);
+        wait_condvar(&queue->value_available, &queue->mutex);
     }
 
-    return rb_ensure(queue_pop_inner, (VALUE)queue,
-		     unlock_mutex_call, (VALUE)&queue->mutex);
-}
-
-static VALUE
-queue_pop_inner(VALUE arg)
-{
-    Queue *queue = (Queue *)arg;
-    VALUE result = shift_list(&queue->values);
+    result = shift_list(&queue->values);
     if (queue->capacity && queue->values.size < queue->capacity) {
         signal_condvar(&queue->space_available);
     }
+    unlock_mutex(&queue->mutex);
+
     return result;
 }
 
@@ -1088,11 +1010,11 @@ rb_queue_push(VALUE self, VALUE value)
 
     lock_mutex(&queue->mutex);
     while (queue->capacity && queue->values.size >= queue->capacity) {
-        wait_queue(&queue->space_available, &queue->mutex);
+        wait_condvar(&queue->space_available, &queue->mutex);
     }
     push_list(&queue->values, value);
-    rb_ensure(signal_condvar_call, (VALUE)&queue->value_available,
-	      unlock_mutex_call, (VALUE)&queue->mutex);
+    signal_condvar(&queue->value_available);
+    unlock_mutex(&queue->mutex);
 
     return self;
 }
