@@ -3,7 +3,7 @@
   object.c -
 
   $Author: shyouhei $
-  $Date: 2008-06-18 00:21:30 -0600 (Wed, 18 Jun 2008) $
+  $Date: 2009-12-14 12:53:19 +0900 (Mon, 14 Dec 2009) $
   created at: Thu Jul 15 12:01:24 JST 1993
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -314,7 +314,7 @@ VALUE
 rb_any_to_s(obj)
     VALUE obj;
 {
-    char *cname = rb_obj_classname(obj);
+    const char *cname = rb_obj_classname(obj);
     size_t len;
     VALUE str;
 
@@ -341,7 +341,7 @@ inspect_i(id, value, str)
     VALUE str;
 {
     VALUE str2;
-    char *ivname;
+    const char *ivname;
 
     /* need not to show internal data */
     if (CLASS_OF(value) == 0) return ST_CONTINUE;
@@ -397,9 +397,8 @@ rb_obj_inspect(obj)
 	&& ROBJECT(obj)->iv_tbl->num_entries > 0) {
 	VALUE str;
 	size_t len;
-	char *c;
+	const char *c = rb_obj_classname(obj);
 
-	c = rb_obj_classname(obj);
 	if (rb_inspecting_p(obj)) {
 	    len = strlen(c)+10+16+1;
 	    str = rb_str_new(0, len); /* 10:tags 16:addr 1:nul */
@@ -494,6 +493,35 @@ rb_obj_is_kind_of(obj, c)
 }
 
 
+/*
+ *  call-seq:
+ *     obj.tap{|x|...}    => obj
+ *  
+ *  Yields <code>x</code> to the block, and then returns <code>x</code>.
+ *  The primary purpose of this method is to "tap into" a method chain,
+ *  in order to perform operations on intermediate results within the chain.
+ *
+ *	(1..10).tap {
+ *	  |x| puts "original: #{x.inspect}"
+ *	}.to_a.tap {
+ *	  |x| puts "array: #{x.inspect}"
+ *	}.select {|x| x%2==0}.tap {
+ *	  |x| puts "evens: #{x.inspect}"
+ *	}.map {|x| x*x}.tap {
+ *	  |x| puts "squares: #{x.inspect}"
+ *	}
+ *
+ */
+
+VALUE
+rb_obj_tap(obj)
+    VALUE obj;
+{
+    rb_yield(obj);
+    return obj;
+}
+
+
 /*
  * Document-method: inherited
  *
@@ -1132,7 +1160,7 @@ sym_inspect(sym)
     VALUE sym;
 {
     VALUE str;
-    char *name;
+    const char *name;
     ID id = SYM2ID(sym);
 
     name = rb_id2name(id);
@@ -1183,6 +1211,37 @@ sym_to_sym(sym)
 }
 
 
+static VALUE
+sym_call(args, mid)
+    VALUE args, mid;
+{
+    VALUE obj;
+
+    if (RARRAY(args)->len < 1) {
+	rb_raise(rb_eArgError, "no receiver given");
+    }
+    obj = rb_ary_shift(args);
+    return rb_apply(obj, (ID)mid, args);
+}
+
+VALUE rb_proc_new _((VALUE (*)(ANYARGS/* VALUE yieldarg[, VALUE procarg] */), VALUE));
+
+/*
+ * call-seq:
+ *   sym.to_proc
+ *
+ * Returns a _Proc_ object which respond to the given method by _sym_.
+ *
+ *   (1..3).collect(&:to_s)  #=> ["1", "2", "3"]
+ */
+
+static VALUE
+sym_to_proc(VALUE sym)
+{
+    return rb_proc_new(sym_call, (VALUE)SYM2ID(sym));
+}
+
+
 /***********************************************************************
  *
  * Document-class: Module
@@ -1514,8 +1573,21 @@ rb_class_initialize(argc, argv, klass)
  *  call-seq:
  *     class.allocate()   =>   obj
  *  
- *  Allocates space for a new object of <i>class</i>'s class. The
- *  returned object must be an instance of <i>class</i>.
+ *  Allocates space for a new object of <i>class</i>'s class and does not
+ *  call initialize on the new instance. The returned object must be an
+ *  instance of <i>class</i>.
+ *  
+ *      klass = Class.new do
+ *        def initialize(*args)
+ *          @initialized = true
+ *        end
+ *      
+ *        def initialized?
+ *          @initialized || false
+ *        end
+ *      end
+ *      
+ *      klass.allocate.initialized? #=> false
  *     
  */
 
@@ -2199,6 +2271,19 @@ rb_to_integer(val, method)
     return v;
 }
 
+VALUE
+rb_check_to_integer(VALUE val, const char *method)
+{
+    VALUE v;
+
+    if (FIXNUM_P(val)) return val;
+    v = convert_type(val, "Integer", method, Qfalse);
+    if (!rb_obj_is_kind_of(v, rb_cInteger)) {
+	return Qnil;
+    }
+    return v;
+}
+
 VALUE
 rb_to_int(val)
     VALUE val;
@@ -2385,13 +2470,8 @@ rb_Float(val)
 	break;
 
       default:
-      {
-	  VALUE f = rb_convert_type(val, T_FLOAT, "Float", "to_f");
-	  if (isnan(RFLOAT(f)->value)) {
-	      rb_raise(rb_eArgError, "invalid value for Float()");
-	  }
-	  return f;
-      }
+	return rb_convert_type(val, T_FLOAT, "Float", "to_f");
+
     }
 }
 
@@ -2672,6 +2752,7 @@ Init_Object()
     rb_define_method(rb_mKernel, "instance_of?", rb_obj_is_instance_of, 1);
     rb_define_method(rb_mKernel, "kind_of?", rb_obj_is_kind_of, 1);
     rb_define_method(rb_mKernel, "is_a?", rb_obj_is_kind_of, 1);
+    rb_define_method(rb_mKernel, "tap", rb_obj_tap, 0);
 
     rb_define_private_method(rb_mKernel, "singleton_method_added", rb_obj_dummy, 1);
     rb_define_private_method(rb_mKernel, "singleton_method_removed", rb_obj_dummy, 1);
@@ -2713,6 +2794,7 @@ Init_Object()
     rb_define_method(rb_cSymbol, "to_s", sym_to_s, 0);
     rb_define_method(rb_cSymbol, "id2name", sym_to_s, 0);
     rb_define_method(rb_cSymbol, "to_sym", sym_to_sym, 0);
+    rb_define_method(rb_cSymbol, "to_proc", sym_to_proc, 0);
     rb_define_method(rb_cSymbol, "===", rb_obj_equal, 1); 
 
     rb_define_method(rb_cModule, "freeze", rb_mod_freeze, 0);
