@@ -3,7 +3,7 @@
   sprintf.c -
 
   $Author: shyouhei $
-  $Date: 2008-06-19 17:12:46 -0600 (Thu, 19 Jun 2008) $
+  $Date: 2008-06-20 08:12:46 +0900 (Fri, 20 Jun 2008) $
   created at: Fri Oct 15 10:39:26 JST 1993
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -727,7 +727,7 @@ rb_str_format(argc, argv, fmt)
 		fval = RFLOAT(rb_Float(val))->value;
 #if defined(_WIN32) && !defined(__BORLANDC__)
 		if (isnan(fval) || isinf(fval)) {
-		    char *expr;
+		    const char *expr;
 
 		    if  (isnan(fval)) {
 			expr = "NaN";
