@@ -21,10 +21,15 @@
 #include <stdlib.h>
 #include <errno.h>
 #include <assert.h>
+#include <ctype.h>
 
 #include <windows.h>
 #include <winbase.h>
 #include <wincon.h>
+#include <shlobj.h>
+#if _MSC_VER >= 1400
+#include <crtdbg.h>
+#endif
 #ifdef __MINGW32__
 #include <mswsock.h>
 #include <mbstring.h>
@@ -173,19 +178,64 @@ static struct {
     {	ERROR_INFLOOP_IN_RELOC_CHAIN,	ENOEXEC		},
     {	ERROR_FILENAME_EXCED_RANGE,	ENOENT		},
     {	ERROR_NESTING_NOT_ALLOWED,	EAGAIN		},
+#ifndef ERROR_PIPE_LOCAL
+#define ERROR_PIPE_LOCAL	229L
+#endif
+    {	ERROR_PIPE_LOCAL,		EPIPE		},
+    {	ERROR_BAD_PIPE,			EPIPE		},
+    {	ERROR_PIPE_BUSY,		EAGAIN		},
+    {	ERROR_NO_DATA,			EPIPE		},
+    {	ERROR_PIPE_NOT_CONNECTED,	EPIPE		},
+    {	ERROR_OPERATION_ABORTED,	EINTR		},
     {	ERROR_NOT_ENOUGH_QUOTA,		ENOMEM		},
-    {	WSAENAMETOOLONG,		ENAMETOOLONG	},
-    {	WSAENOTEMPTY,			ENOTEMPTY	},
+    {	ERROR_MOD_NOT_FOUND,		ENOENT		},
     {	WSAEINTR,			EINTR		},
     {	WSAEBADF,			EBADF		},
     {	WSAEACCES,			EACCES		},
     {	WSAEFAULT,			EFAULT		},
     {	WSAEINVAL,			EINVAL		},
     {	WSAEMFILE,			EMFILE		},
+    {	WSAEWOULDBLOCK,			EWOULDBLOCK	},
+    {	WSAEINPROGRESS,			EINPROGRESS	},
+    {	WSAEALREADY,			EALREADY	},
+    {	WSAENOTSOCK,			ENOTSOCK	},
+    {	WSAEDESTADDRREQ,		EDESTADDRREQ	},
+    {	WSAEMSGSIZE,			EMSGSIZE	},
+    {	WSAEPROTOTYPE,			EPROTOTYPE	},
+    {	WSAENOPROTOOPT,			ENOPROTOOPT	},
+    {	WSAEPROTONOSUPPORT,		EPROTONOSUPPORT	},
+    {	WSAESOCKTNOSUPPORT,		ESOCKTNOSUPPORT	},
+    {	WSAEOPNOTSUPP,			EOPNOTSUPP	},
+    {	WSAEPFNOSUPPORT,		EPFNOSUPPORT	},
+    {	WSAEAFNOSUPPORT,		EAFNOSUPPORT	},
+    {	WSAEADDRINUSE,			EADDRINUSE	},
+    {	WSAEADDRNOTAVAIL,		EADDRNOTAVAIL	},
+    {	WSAENETDOWN,			ENETDOWN	},
+    {	WSAENETUNREACH,			ENETUNREACH	},
+    {	WSAENETRESET,			ENETRESET	},
+    {	WSAECONNABORTED,		ECONNABORTED	},
+    {	WSAECONNRESET,			ECONNRESET	},
+    {	WSAENOBUFS,			ENOBUFS		},
+    {	WSAEISCONN,			EISCONN		},
+    {	WSAENOTCONN,			ENOTCONN	},
+    {	WSAESHUTDOWN,			ESHUTDOWN	},
+    {	WSAETOOMANYREFS,		ETOOMANYREFS	},
+    {	WSAETIMEDOUT,			ETIMEDOUT	},
+    {	WSAECONNREFUSED,		ECONNREFUSED	},
+    {	WSAELOOP,			ELOOP		},
+    {	WSAENAMETOOLONG,		ENAMETOOLONG	},
+    {	WSAEHOSTDOWN,			EHOSTDOWN	},
+    {	WSAEHOSTUNREACH,		EHOSTUNREACH	},
+    {	WSAEPROCLIM,			EPROCLIM	},
+    {	WSAENOTEMPTY,			ENOTEMPTY	},
+    {	WSAEUSERS,			EUSERS		},
+    {	WSAEDQUOT,			EDQUOT		},
+    {	WSAESTALE,			ESTALE		},
+    {	WSAEREMOTE,			EREMOTE		},
 };
 
-static int
-map_errno(DWORD winerr)
+int
+rb_w32_map_errno(DWORD winerr)
 {
     int i;
 
@@ -205,7 +255,9 @@ map_errno(DWORD winerr)
     return EINVAL;
 }
 
-static char *NTLoginName;
+#define map_errno rb_w32_map_errno
+
+static const char *NTLoginName;
 
 #ifdef WIN95
 static DWORD Win32System = (DWORD)-1;
@@ -362,12 +414,20 @@ flock(int fd, int oper)
 static void init_stdhandle(void);
 
 #if _MSC_VER >= 1400
-static void invalid_parameter(const wchar_t *expr, const wchar_t *func, const wchar_t *file, unsigned int line, uintptr_t dummy)
+static void
+invalid_parameter(const wchar_t *expr, const wchar_t *func, const wchar_t *file, unsigned int line, uintptr_t dummy)
 {
     // nothing to do
 }
+
+static int __cdecl
+rtc_error_handler(int e, const char *src, int line, const char *exe, const char *fmt, ...)
+{
+    return 0;
+}
 #endif
 
+static CRITICAL_SECTION select_mutex;
 static BOOL fWinsock;
 static char *envarea;
 static void
@@ -381,6 +441,61 @@ exit_handler(void)
 	FreeEnvironmentStrings(envarea);
 	envarea = NULL;
     }
+    DeleteCriticalSection(&select_mutex);
+}
+
+static void
+init_env(void)
+{
+    char env[_MAX_PATH];
+    DWORD len;
+    BOOL f;
+    LPITEMIDLIST pidl;
+
+    if (!GetEnvironmentVariable("HOME", env, sizeof(env))) {
+	f = FALSE;
+	if (GetEnvironmentVariable("HOMEDRIVE", env, sizeof(env)))
+	    len = strlen(env);
+	else
+	    len = 0;
+	if (GetEnvironmentVariable("HOMEPATH", env + len, sizeof(env) - len) || len) {
+	    f = TRUE;
+	}
+	else if (GetEnvironmentVariable("USERPROFILE", env, sizeof(env))) {
+	    f = TRUE;
+	}
+	else if (SHGetSpecialFolderLocation(NULL, CSIDL_PERSONAL, &pidl) == 0) {
+	    LPMALLOC alloc;
+	    f = SHGetPathFromIDList(pidl, env);
+	    SHGetMalloc(&alloc);
+	    alloc->lpVtbl->Free(alloc, pidl);
+	    alloc->lpVtbl->Release(alloc);
+	}
+	if (f) {
+	    char *p = env;
+	    while (*p) {
+		if (*p == '\\') *p = '/';
+		p = CharNext(p);
+	    }
+	    if (p - env == 2 && env[1] == ':') {
+		*p++ = '/';
+		*p = 0;
+	    }
+	    SetEnvironmentVariable("HOME", env);
+	}
+    }
+
+    if (!GetEnvironmentVariable("USER", env, sizeof env)) {
+	if (GetEnvironmentVariable("USERNAME", env, sizeof env) ||
+	    GetUserName(env, (len = sizeof env, &len))) {
+	    SetEnvironmentVariable("USER", env);
+	}
+	else {
+	    NTLoginName = "<Unknown>";
+	    return;
+	}
+    }
+    NTLoginName = strdup(env);
 }
 
 //
@@ -389,14 +504,12 @@ exit_handler(void)
 void
 NtInitialize(int *argc, char ***argv)
 {
-
-    WORD version;
-    int ret;
-
 #if _MSC_VER >= 1400
     static void set_pioinfo_extra(void);
 
+    _CrtSetReportMode(_CRT_ASSERT, 0);
     _set_invalid_parameter_handler(invalid_parameter);
+    _RTC_SetErrorFunc(rtc_error_handler);
     set_pioinfo_extra();
 #endif
 
@@ -411,8 +524,12 @@ NtInitialize(int *argc, char ***argv)
 
     tzset();
 
+    init_env();
+
     init_stdhandle();
 
+    InitializeCriticalSection(&select_mutex);
+
     atexit(exit_handler);
 
     // Initialize Winsock
@@ -427,22 +544,7 @@ NtInitialize(int *argc, char ***argv)
 char *
 getlogin()
 {
-    char buffer[200];
-    DWORD len = 200;
-    extern char *NTLoginName;
-
-    if (NTLoginName == NULL) {
-	if (GetUserName(buffer, &len)) {
-	    NTLoginName = (char *)malloc(len+1);
-	    if (!NTLoginName) return NULL;
-	    strncpy(NTLoginName, buffer, len);
-	    NTLoginName[len] = '\0';
-	}
-	else {
-	    NTLoginName = "<Unknown>";
-	}
-    }
-    return NTLoginName;
+    return (char *)NTLoginName;
 }
 
 #define MAXCHILDNUM 256	/* max num of child processes */
@@ -457,15 +559,6 @@ static struct ChildRecord {
     for (v = ChildRecord; v < ChildRecord + sizeof(ChildRecord) / sizeof(ChildRecord[0]); ++v)
 #define END_FOREACH_CHILD } while (0)
 
-static struct ChildRecord *
-FindFirstChildSlot(void)
-{
-    FOREACH_CHILD(child) {
-	if (child->pid) return child;
-    } END_FOREACH_CHILD;
-    return NULL;
-}
-
 static struct ChildRecord *
 FindChildSlot(rb_pid_t pid)
 {
@@ -478,6 +571,18 @@ FindChildSlot(rb_pid_t pid)
     return NULL;
 }
 
+static struct ChildRecord *
+FindChildSlotByHandle(HANDLE h)
+{
+
+    FOREACH_CHILD(child) {
+	if (child->hProcess == h) {
+	    return child;
+	}
+    } END_FOREACH_CHILD;
+    return NULL;
+}
+
 static void
 CloseChildHandle(struct ChildRecord *child)
 {
@@ -617,7 +722,7 @@ rb_w32_get_osfhandle(int fh)
 }
 
 rb_pid_t
-pipe_exec(char *cmd, int mode, FILE **fpr, FILE **fpw)
+pipe_exec(const char *cmd, int mode, FILE **fpr, FILE **fpw)
 {
     struct ChildRecord* child;
     HANDLE hReadIn, hReadOut;
@@ -756,7 +861,7 @@ pipe_exec(char *cmd, int mode, FILE **fpr, FILE **fpw)
 extern VALUE rb_last_status;
 
 int
-do_spawn(int mode, char *cmd)
+do_spawn(int mode, const char *cmd)
 {
     struct ChildRecord *child;
     DWORD exitcode;
@@ -793,7 +898,7 @@ do_spawn(int mode, char *cmd)
 }
 
 int
-do_aspawn(int mode, char *prog, char **argv)
+do_aspawn(int mode, const char *prog, char **argv)
 {
     char *cmd, *p, *q, *s, **t;
     int len, n, bs, quote;
@@ -1006,11 +1111,13 @@ CreateChild(const char *cmd, const char *prog, SECURITY_ATTRIBUTES *psa,
 	}
     }
     if (p) {
-	shell = p;
-	while (*p) {
-	    if ((unsigned char)*p == '/')
-		*p = '\\';
-	    p = CharNext(p);
+	char *tmp = ALLOCA_N(char, strlen(p) + 1);
+	strcpy(tmp, p);
+	shell = tmp;
+	while (*tmp) {
+	    if ((unsigned char)*tmp == '/')
+		*tmp = '\\';
+	    tmp = CharNext(tmp);
 	}
     }
 
@@ -1064,10 +1171,9 @@ insert(const char *path, VALUE vinfo)
     if (!tmpcurr) return -1;
     MEMZERO(tmpcurr, NtCmdLineElement, 1);
     tmpcurr->len = strlen(path);
-    tmpcurr->str = (char *)malloc(tmpcurr->len + 1);
+    tmpcurr->str = strdup(path);
     if (!tmpcurr->str) return -1;
     tmpcurr->flags |= NTMALLOC;
-    strcpy(tmpcurr->str, path);
     **tail = tmpcurr;
     *tail = &tmpcurr->next;
 
@@ -1164,7 +1270,7 @@ skipspace(char *ptr)
 int 
 rb_w32_cmdvector(const char *cmd, char ***vec)
 {
-    int cmdlen, globbing, len, i;
+    int globbing, len;
     int elements, strsz, done;
     int slashes, escape;
     char *ptr, *base, *buffer, *cmdline;
@@ -1250,10 +1356,13 @@ rb_w32_cmdvector(const char *cmd, char ***vec)
 		if (!(slashes & 1)) {
 		    if (!quote)
 			quote = *ptr;
-		    else if (quote == *ptr)
+		    else if (quote == *ptr) {
+			if (quote == '"' && quote == ptr[1])
+			    ptr++;
 			quote = '\0';
-		    escape++;
+		    }
 		}
+		escape++;
 		slashes = 0;
 		break;
 
@@ -1280,10 +1389,10 @@ rb_w32_cmdvector(const char *cmd, char ***vec)
 	//
 
 	if (escape) {
-	    char *p = base;
+	    char *p = base, c;
 	    slashes = quote = 0;
 	    while (p < base + len) {
-		switch (*p) {
+		switch (c = *p) {
 		  case '\\':
 		    p++;
 		    if (quote != '\'') slashes++;
@@ -1291,33 +1400,27 @@ rb_w32_cmdvector(const char *cmd, char ***vec)
 
 		  case '\'':
 		  case '"':
-		    if (!(slashes & 1)) {
-			if (!quote)
-			    quote = *p;
-			else if (quote == *p)
-			    quote = '\0';
-			else {
-			    p++;
-			    slashes = 0;
-			    break;
-			}
-		    }
-		    if (base + slashes == p) {
-			base += slashes >> 1;
-			len -= slashes >> 1;
-			slashes &= 1;
-		    }
-		    if (base == p) {
-			base = ++p;
-			--len;
-		    }
-		    else {
-			memcpy(p - ((slashes + 1) >> 1), p + (~slashes & 1), base + len - p);
-			slashes >>= 1;
-			p -= slashes;
-			len -= slashes + 1;
+		    if (!(slashes & 1) && quote && quote != c) {
+			p++;
 			slashes = 0;
+			break;
 		    }
+		    memcpy(p - ((slashes + 1) >> 1), p + (~slashes & 1),
+			   base + len - p);
+		    len -= ((slashes + 1) >> 1) + (~slashes & 1);
+		    p -= (slashes + 1) >> 1;
+		    if (!(slashes & 1)) {
+			if (quote) {
+			    if (quote == '"' && quote == *p)
+				p++;
+			    quote = '\0';
+			}
+			else
+			    quote = c;
+		    }
+		    else
+			p++;
+		    slashes = 0;
 		    break;
 
 		  default:
@@ -1384,10 +1487,10 @@ rb_w32_cmdvector(const char *cmd, char ***vec)
     ptr = buffer + (elements+1) * sizeof(char *);
 
     while (curr = cmdhead) {
-	strncpy (ptr, curr->str, curr->len);
-	ptr[curr->len] = '\0';
+	memcpy(ptr, curr->str, curr->len);
 	*vptr++ = ptr;
-	ptr += curr->len + 1;
+	ptr += curr->len;
+	*ptr++ = '\0';
 	cmdhead = curr->next;
 	if (curr->flags & NTMALLOC) free(curr->str);
 	free(curr);
@@ -1418,14 +1521,44 @@ rb_w32_cmdvector(const char *cmd, char ***vec)
 #define BitOfIsRep(n) ((n) * 2 + 1)
 #define DIRENT_PER_CHAR (CHAR_BIT / 2)
 
+static HANDLE
+open_dir_handle(const char *filename, WIN32_FIND_DATA *fd)
+{
+    HANDLE fh;
+    static const char wildcard[] = "/*";
+    long len = strlen(filename);
+    char *scanname = malloc(len + sizeof(wildcard));
+
+    //
+    // Create the search pattern
+    //
+    if (!scanname) {
+	return INVALID_HANDLE_VALUE;
+    }
+    memcpy(scanname, filename, len + 1);
+
+    if (index("/\\:", *CharPrev(scanname, scanname + len)) == NULL)
+	memcpy(scanname + len, wildcard, sizeof(wildcard));
+    else
+	memcpy(scanname + len, wildcard + 1, sizeof(wildcard) - 1);
+
+    //
+    // do the FindFirstFile call
+    //
+    fh = FindFirstFile(scanname, fd);
+    free(scanname);
+    if (fh == INVALID_HANDLE_VALUE) {
+	errno = map_errno(GetLastError());
+    }
+    return fh;
+}
+
 DIR *
 rb_w32_opendir(const char *filename)
 {
     DIR               *p;
     long               len;
     long               idx;
-    char               scannamespc[PATHLEN];
-    char	      *scanname = scannamespc;
     struct stat	       sbuf;
     WIN32_FIND_DATA fd;
     HANDLE          fh;
@@ -1443,35 +1576,17 @@ rb_w32_opendir(const char *filename)
 	return NULL;
     }
 
+    fh = open_dir_handle(filename, &fd);
+    if (fh == INVALID_HANDLE_VALUE) {
+	return NULL;
+    }
+
     //
     // Get us a DIR structure
     //
-
-    p = xcalloc(sizeof(DIR), 1);
+    p = calloc(sizeof(DIR), 1);
     if (p == NULL)
 	return NULL;
-    
-    //
-    // Create the search pattern
-    //
-
-    strcpy(scanname, filename);
-
-    if (index("/\\:", *CharPrev(scanname, scanname + strlen(scanname))) == NULL)
-	strcat(scanname, "/*");
-    else
-	strcat(scanname, "*");
-
-    //
-    // do the FindFirstFile call
-    //
-
-    fh = FindFirstFile(scanname, &fd);
-    if (fh == INVALID_HANDLE_VALUE) {
-	errno = map_errno(GetLastError());
-	free(p);
-	return NULL;
-    }
 
     //
     // now allocate the first part of the string table for the
@@ -1603,15 +1718,7 @@ rb_w32_readdir(DIR *dirp)
 long
 rb_w32_telldir(DIR *dirp)
 {
-    long loc = 0; char *p = dirp->curr;
-
-    rb_w32_rewinddir(dirp);
-
-    while (p != dirp->curr) {
-	move_to_next_entry(dirp); loc++;
-    }
-
-    return loc;
+    return dirp->loc;
 }
 
 //
@@ -1636,6 +1743,7 @@ void
 rb_w32_rewinddir(DIR *dirp)
 {
     dirp->curr = dirp->start;
+    dirp->loc = 0;
 }
 
 //
@@ -1870,7 +1978,7 @@ rb_w32_strerror(int e)
 	    e = GetLastError();
 	if (FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM |
 			  FORMAT_MESSAGE_IGNORE_INSERTS, &source, e, 0,
-			  buffer, 512, NULL) == 0) {
+			  buffer, sizeof(buffer), NULL) == 0) {
 	    strcpy(buffer, "Unknown Error");
 	}
     }
@@ -2017,87 +2125,275 @@ rb_w32_fdisset(int fd, fd_set *set)
 static int NtSocketsInitialized = 0;
 
 static int
-extract_file_fd(fd_set *set, fd_set *fileset)
+extract_fd(fd_set *dst, fd_set *src, int (*func)(SOCKET))
 {
-    int idx;
+    int s = 0;
+    if (!src || !dst) return 0;
 
-    fileset->fd_count = 0;
-    if (!set)
-	return 0;
-    for (idx = 0; idx < set->fd_count; idx++) {
-	SOCKET fd = set->fd_array[idx];
+    while (s < src->fd_count) {
+        SOCKET fd = src->fd_array[s];
 
-	if (!is_socket(fd)) {
-	    int i;
+	if (!func || (*func)(fd)) { /* move it to dst */
+	    int d;
 
-	    for (i = 0; i < fileset->fd_count; i++) {
-		if (fileset->fd_array[i] == fd) {
-		    break;
-		}
+	    for (d = 0; d < dst->fd_count; d++) {
+		if (dst->fd_array[d] == fd) break;
 	    }
-	    if (i == fileset->fd_count) {
-		if (fileset->fd_count < FD_SETSIZE) {
-		    fileset->fd_array[i] = fd;
-		    fileset->fd_count++;
-		}
+	    if (d == dst->fd_count && dst->fd_count < FD_SETSIZE) {
+		dst->fd_array[dst->fd_count++] = fd;
 	    }
+	    memmove(
+		&src->fd_array[s],
+		&src->fd_array[s+1], 
+		sizeof(src->fd_array[0]) * (--src->fd_count - s));
 	}
+	else s++;
     }
-    return fileset->fd_count;
+
+    return dst->fd_count;
 }
 
+static int
+is_not_socket(SOCKET sock)
+{
+    return !is_socket(sock);
+}
+
+static int
+is_pipe(SOCKET sock) /* DONT call this for SOCKET! it clains it is PIPE. */
+{
+    int ret;
+
+    RUBY_CRITICAL(
+	ret = (GetFileType((HANDLE)sock) == FILE_TYPE_PIPE)
+    );
+
+    return ret;
+}
+
+static int
+is_readable_pipe(SOCKET sock) /* call this for pipe only */
+{
+    int ret;
+    DWORD n = 0;
+
+    RUBY_CRITICAL(
+	if (PeekNamedPipe((HANDLE)sock, NULL, 0, NULL, &n, NULL)) {
+	    ret = (n > 0);
+	}
+	else {
+	    ret = (GetLastError() == ERROR_BROKEN_PIPE); /* pipe was closed */
+	}
+    );
+
+    return ret;
+}
+
+static int
+is_console(SOCKET sock) /* DONT call this for SOCKET! */
+{
+    int ret;
+    DWORD n = 0;
+    INPUT_RECORD ir;
+
+    RUBY_CRITICAL(
+	ret = (PeekConsoleInput((HANDLE)sock, &ir, 1, &n))
+    );
+
+    return ret;
+}
+
+static int
+is_readable_console(SOCKET sock) /* call this for console only */
+{
+    int ret = 0;
+    DWORD n = 0;
+    INPUT_RECORD ir;
+
+    RUBY_CRITICAL(
+	if (PeekConsoleInput((HANDLE)sock, &ir, 1, &n) && n > 0) {
+	    if (ir.EventType == KEY_EVENT && ir.Event.KeyEvent.bKeyDown &&
+		ir.Event.KeyEvent.uChar.AsciiChar) {
+		ret = 1;
+	    }
+	    else {
+		ReadConsoleInput((HANDLE)sock, &ir, 1, &n);
+	    }
+	}
+    );
+
+    return ret;
+}
+
+static int
+do_select(int nfds, fd_set *rd, fd_set *wr, fd_set *ex,
+            struct timeval *timeout)
+{
+    int r = 0;
+
+    if (nfds == 0) {
+	if (timeout)
+	    rb_w32_sleep(timeout->tv_sec * 1000 + timeout->tv_usec / 1000);
+	else
+	    rb_w32_sleep(INFINITE);
+    }
+    else {
+	RUBY_CRITICAL(
+	    EnterCriticalSection(&select_mutex);
+	    r = select(nfds, rd, wr, ex, timeout);
+	    LeaveCriticalSection(&select_mutex);
+	    if (r == SOCKET_ERROR) {
+		errno = map_errno(WSAGetLastError());
+		r = -1;
+	    }
+	);
+    }
+
+    return r;
+}
+
+static inline int
+subtract(struct timeval *rest, const struct timeval *wait)
+{
+    if (rest->tv_sec < wait->tv_sec) {
+	return 0;
+    }
+    while (rest->tv_usec <= wait->tv_usec) {
+	if (rest->tv_sec <= wait->tv_sec) {
+	    return 0;
+	}
+	rest->tv_sec -= 1;
+	rest->tv_usec += 1000 * 1000;
+    }
+    rest->tv_sec -= wait->tv_sec;
+    rest->tv_usec -= wait->tv_usec;
+    return 1;
+}
+
+static inline int
+compare(const struct timeval *t1, const struct timeval *t2)
+{
+    if (t1->tv_sec < t2->tv_sec)
+	return -1;
+    if (t1->tv_sec > t2->tv_sec)
+	return 1;
+    if (t1->tv_usec < t2->tv_usec)
+	return -1;
+    if (t1->tv_usec > t2->tv_usec)
+	return 1;
+    return 0;
+}
+
+#undef Sleep
 long 
-rb_w32_select (int nfds, fd_set *rd, fd_set *wr, fd_set *ex,
-	       struct timeval *timeout)
+rb_w32_select(int nfds, fd_set *rd, fd_set *wr, fd_set *ex,
+	      struct timeval *timeout)
 {
     long r;
-    fd_set file_rd;
-    fd_set file_wr;
-#ifdef USE_INTERRUPT_WINSOCK
-    fd_set trap;
-#endif /* USE_INTERRUPT_WINSOCK */
-    int file_nfds;
+    fd_set pipe_rd;
+    fd_set cons_rd;
+    fd_set else_rd;
+    fd_set else_wr;
+    int nonsock = 0;
+    struct timeval limit;
+
+    if (nfds < 0 || (timeout && (timeout->tv_sec < 0 || timeout->tv_usec < 0))) {
+	errno = EINVAL;
+	return -1;
+    }
+
+    if (timeout) {
+	if (timeout->tv_sec < 0 ||
+	    timeout->tv_usec < 0 ||
+	    timeout->tv_usec >= 1000000) {
+	    errno = EINVAL;
+	    return -1;
+	}
+	gettimeofday(&limit, NULL);
+	limit.tv_sec += timeout->tv_sec;
+	limit.tv_usec += timeout->tv_usec;
+	if (limit.tv_usec >= 1000000) {
+	    limit.tv_usec -= 1000000;
+	    limit.tv_sec++;
+	}
+    }
 
     if (!NtSocketsInitialized) {
 	StartSockets();
     }
+
+    // assume else_{rd,wr} (other than socket, pipe reader, console reader)
+    // are always readable/writable. but this implementation still has
+    // problem. if pipe's buffer is full, writing to pipe will block
+    // until some data is read from pipe. but ruby is single threaded system,
+    // so whole system will be blocked forever.
+
+    else_rd.fd_count = 0;
+    nonsock += extract_fd(&else_rd, rd, is_not_socket);
+
+    pipe_rd.fd_count = 0;
+    extract_fd(&pipe_rd, &else_rd, is_pipe); // should not call is_pipe for socket
+
+    cons_rd.fd_count = 0;
+    extract_fd(&cons_rd, &else_rd, is_console); // ditto
+
+    else_wr.fd_count = 0;
+    nonsock += extract_fd(&else_wr, wr, is_not_socket);
+
     r = 0;
     if (rd && rd->fd_count > r) r = rd->fd_count;
     if (wr && wr->fd_count > r) r = wr->fd_count;
     if (ex && ex->fd_count > r) r = ex->fd_count;
     if (nfds > r) nfds = r;
-    if (nfds == 0 && timeout) {
-	Sleep(timeout->tv_sec * 1000 + timeout->tv_usec / 1000);
-	return 0;
-    }
-    file_nfds = extract_file_fd(rd, &file_rd);
-    file_nfds += extract_file_fd(wr, &file_wr);
-    if (file_nfds)
+
     {
-	// assume normal files are always readable/writable
-	// fake read/write fd_set and return value
-	if (rd) *rd = file_rd;
-	if (wr) *wr = file_wr;
-	return file_nfds;
+	struct timeval rest;
+	struct timeval wait;
+	struct timeval zero;
+	wait.tv_sec = 0; wait.tv_usec = 10 * 1000; // 10ms
+	zero.tv_sec = 0; zero.tv_usec = 0;         //  0ms
+	if (timeout) rest = *timeout;
+	for (;;) {
+	    if (nonsock) {
+		// modifying {else,pipe,cons}_rd is safe because
+		// if they are modified, function returns immediately.
+		extract_fd(&else_rd, &pipe_rd, is_readable_pipe);
+		extract_fd(&else_rd, &cons_rd, is_readable_console);
+	    }
+
+	    if (else_rd.fd_count || else_wr.fd_count) {
+		r = do_select(nfds, rd, wr, ex, &zero); // polling
+		if (r < 0) break; // XXX: should I ignore error and return signaled handles?
+		r += extract_fd(rd, &else_rd, NULL); // move all
+		r += extract_fd(wr, &else_wr, NULL); // move all
+		break;
+	    }
+	    else {
+		fd_set orig_rd;
+		fd_set orig_wr;
+		fd_set orig_ex;
+		struct timeval *dowait = &wait;
+		if (timeout && compare(&rest, &wait) < 0) dowait = &rest;
+
+		if (rd) orig_rd = *rd;
+		if (wr) orig_wr = *wr;
+		if (ex) orig_ex = *ex;
+		r = do_select(nfds, rd, wr, ex, dowait);
+		if (r != 0) break; // signaled or error
+		if (rd) *rd = orig_rd;
+		if (wr) *wr = orig_wr;
+		if (ex) *ex = orig_ex;
+
+		if (timeout) {
+		    struct timeval now;
+		    gettimeofday(&now, NULL);
+		    rest = limit;
+		    if (!subtract(&rest, &now)) break;
+		}
+	    }
+	}
     }
 
-#if USE_INTERRUPT_WINSOCK
-    if (ex)
-	trap = *ex;
-    else
-	trap.fd_count = 0;
-    if (trap.fd_count < FD_SETSIZE)
-	trap.fd_array[trap.fd_count++] = (SOCKET)interrupted_event;
-    // else unable to catch interrupt.
-    ex = &trap;
-#endif /* USE_INTERRUPT_WINSOCK */
-
-    RUBY_CRITICAL({
-	r = select(nfds, rd, wr, ex, timeout);
-	if (r == SOCKET_ERROR) {
-	    errno = map_errno(WSAGetLastError());
-	}
-    });
     return r;
 }
 
@@ -2166,21 +2462,32 @@ int
 rb_w32_accept(int s, struct sockaddr *addr, int *addrlen)
 {
     SOCKET r;
+    int fd;
 
     if (!NtSocketsInitialized) {
 	StartSockets();
     }
     RUBY_CRITICAL({
-	r = accept(TO_SOCKET(s), addr, addrlen);
-	if (r == INVALID_SOCKET) {
-	    errno = map_errno(WSAGetLastError());
-	    s = -1;
-	}
-	else {
-	    s = rb_w32_open_osfhandle(r, O_RDWR|O_BINARY);
+	HANDLE h = CreateFile("NUL", 0, 0, NULL, OPEN_ALWAYS, 0, NULL);
+	fd = rb_w32_open_osfhandle((long)h, O_RDWR|O_BINARY|O_NOINHERIT);
+	if (fd != -1) {
+	    r = accept(TO_SOCKET(s), addr, addrlen);
+	    if (r != INVALID_SOCKET) {
+		MTHREAD_ONLY(EnterCriticalSection(&(_pioinfo(fd)->lock)));
+		_set_osfhnd(fd, r);
+		MTHREAD_ONLY(LeaveCriticalSection(&_pioinfo(fd)->lock));
+		CloseHandle(h);
+	    }
+	    else {
+		errno = map_errno(WSAGetLastError());
+		close(fd);
+		fd = -1;
+	    }
 	}
+	else
+	    CloseHandle(h);
     });
-    return s;
+    return fd;
 }
 
 #undef bind
@@ -2213,14 +2520,11 @@ rb_w32_connect(int s, struct sockaddr *addr, int addrlen)
     RUBY_CRITICAL({
 	r = connect(TO_SOCKET(s), addr, addrlen);
 	if (r == SOCKET_ERROR) {
-	    r = WSAGetLastError();
-	    if (r != WSAEWOULDBLOCK) {
-		errno = map_errno(r);
-	    }
-	    else {
+	    int err = WSAGetLastError();
+	    if (err != WSAEWOULDBLOCK)
+		errno = map_errno(err);
+	    else
 		errno = EINPROGRESS;
-		r = -1;
-	    }
 	}
     });
     return r;
@@ -2450,6 +2754,8 @@ open_ifs_socket(int af, int type, int protocol)
 		    out = WSASocket(af, type, protocol, &(proto_buffers[i]), 0, 0);
 		    break;
 		}
+		if (out == INVALID_SOCKET)
+		    out = WSASocket(af, type, protocol, NULL, 0, 0);
 	    }
 
 	    free(proto_buffers);
@@ -2741,7 +3047,7 @@ waitpid(rb_pid_t pid, int *stat_loc, int options)
 	    return -1;
 	}
 
-	return poll_child_status(ChildRecord + ret, stat_loc);
+	return poll_child_status(FindChildSlotByHandle(events[ret]), stat_loc);
     }
     else {
 	struct ChildRecord* child = FindChildSlot(pid);
@@ -2765,24 +3071,36 @@ waitpid(rb_pid_t pid, int *stat_loc, int options)
 
 #include <sys/timeb.h>
 
+static int
+filetime_to_timeval(const FILETIME* ft, struct timeval *tv)
+{
+    ULARGE_INTEGER tmp;
+    unsigned LONG_LONG lt;
+
+    tmp.LowPart = ft->dwLowDateTime;
+    tmp.HighPart = ft->dwHighDateTime;
+    lt = tmp.QuadPart;
+
+    /* lt is now 100-nanosec intervals since 1601/01/01 00:00:00 UTC,
+       convert it into UNIX time (since 1970/01/01 00:00:00 UTC).
+       the first leap second is at 1972/06/30, so we doesn't need to think
+       about it. */
+    lt /= 10;	/* to usec */
+    lt -= (LONG_LONG)((1970-1601)*365.2425) * 24 * 60 * 60 * 1000 * 1000;
+
+    tv->tv_sec = lt / (1000 * 1000);
+    tv->tv_usec = lt % (1000 * 1000);
+
+    return tv->tv_sec > 0 ? 0 : -1;
+}
+
 int _cdecl
 gettimeofday(struct timeval *tv, struct timezone *tz)
 {
-    SYSTEMTIME st;
-    time_t t;
-    struct tm tm;
+    FILETIME ft;
 
-    GetLocalTime(&st);
-    tm.tm_sec = st.wSecond;
-    tm.tm_min = st.wMinute;
-    tm.tm_hour = st.wHour;
-    tm.tm_mday = st.wDay;
-    tm.tm_mon = st.wMonth - 1;
-    tm.tm_year = st.wYear - 1900;
-    tm.tm_isdst = -1;
-    t = mktime(&tm);
-    tv->tv_sec = t;
-    tv->tv_usec = st.wMilliseconds * 1000;
+    GetSystemTimeAsFileTime(&ft);
+    filetime_to_timeval(&ft, tv);
 
     return 0;
 }
@@ -3057,27 +3375,12 @@ isUNCRoot(const char *path)
 static time_t
 filetime_to_unixtime(const FILETIME *ft)
 {
-    FILETIME loc;
-    SYSTEMTIME st;
-    struct tm tm;
-    time_t t;
+    struct timeval tv;
 
-    if (!FileTimeToLocalFileTime(ft, &loc)) {
+    if (filetime_to_timeval(ft, &tv) == (time_t)-1)
 	return 0;
-    }
-    if (!FileTimeToSystemTime(&loc, &st)) {
-	return 0;
-    }
-    memset(&tm, 0, sizeof(tm));
-    tm.tm_year = st.wYear - 1900;
-    tm.tm_mon = st.wMonth - 1;
-    tm.tm_mday = st.wDay;
-    tm.tm_hour = st.wHour;
-    tm.tm_min = st.wMinute;
-    tm.tm_sec = st.wSecond;
-    tm.tm_isdst = -1;
-    t = mktime(&tm);
-    return t == -1 ? 0 : t;
+    else
+	return tv.tv_sec;
 }
 
 static unsigned
@@ -3121,6 +3424,17 @@ fileattr_to_unixmode(DWORD attr, const char *path)
     return mode;
 }
 
+static int
+check_valid_dir(const char *path)
+{
+    WIN32_FIND_DATA fd;
+    HANDLE fh = open_dir_handle(path, &fd);
+    if (fh == INVALID_HANDLE_VALUE)
+	return -1;
+    FindClose(fh);
+    return 0;
+}
+
 static int
 winnt_stat(const char *path, struct stat *st)
 {
@@ -3151,6 +3465,9 @@ winnt_stat(const char *path, struct stat *st)
 	    errno = map_errno(GetLastError());
 	    return -1;
 	}
+	if (attr & FILE_ATTRIBUTE_DIRECTORY) {
+	    if (check_valid_dir(path)) return -1;
+	}
 	st->st_mode  = fileattr_to_unixmode(attr, path);
     }
 
@@ -3160,6 +3477,21 @@ winnt_stat(const char *path, struct stat *st)
     return 0;
 }
 
+#ifdef WIN95
+static int
+win95_stat(const char *path, struct stat *st)
+{
+    int ret = stat(path, st);
+    if (ret) return ret;
+    if (st->st_mode & S_IFDIR) {
+	return check_valid_dir(path);
+    }
+    return 0;
+}
+#else
+#define win95_stat(path, st) -1
+#endif
+
 int
 rb_w32_stat(const char *path, struct stat *st)
 {
@@ -3195,7 +3527,7 @@ rb_w32_stat(const char *path, struct stat *st)
     } else if (*end == '\\' || (buf1 + 1 == end && *end == ':'))
 	strcat(buf1, ".");
 
-    ret = IsWinNT() ? winnt_stat(buf1, st) : stat(buf1, st);
+    ret = IsWinNT() ? winnt_stat(buf1, st) : win95_stat(buf1, st);
     if (ret == 0) {
 	st->st_mode &= ~(S_IWGRP | S_IWOTH);
     }
@@ -3232,7 +3564,6 @@ rb_w32_times(struct tms *tmbuf)
     return 0;
 }
 
-#undef Sleep
 #define yield_once() Sleep(0)
 #define yield_until(condition) do yield_once(); while (!(condition))
 
@@ -3730,9 +4061,7 @@ int
 rb_w32_utime(const char *path, struct utimbuf *times)
 {
     HANDLE hFile;
-    SYSTEMTIME st;
     FILETIME atime, mtime;
-    struct tm *tm;
     struct stat stat;
     int ret = 0;
 
@@ -3859,6 +4188,10 @@ rb_w32_unlink(const char *path)
 int
 rb_w32_isatty(int fd)
 {
+    // validate fd by using _get_osfhandle() because we cannot access _nhandle
+    if (_get_osfhandle(fd) == -1) {
+	return 0;
+    }
     if (!(_osfile(fd) & FOPEN)) {
 	errno = EBADF;
 	return 0;
@@ -3922,3 +4255,5 @@ rb_w32_fsopen(const char *path, const char *mode, int shflags)
     return f;
 }
 #endif
+
+RUBY_EXTERN int __cdecl _CrtDbgReportW() {return 0;}
